@@ -13,3 +13,7 @@ class Aggregator(Ranker):
 
     def score(self, q: np.ndarray, vectors: np.ndarray, prev_scores: Iterable[float]) -> List[float]:
         return list(prev_scores)
+
+    @property
+    def volume_request(self) -> Optional[int]:
+        return 0
