@@ -10,7 +10,7 @@ from typing import Dict
 import zmq
 from loguru import logger
 from functools import partial
-from concurrent.futures import ThreadPoolExecutor
+import os
 
 SOCK_ADDR = 4
 FAMILY = 0
@@ -42,29 +42,37 @@ class SocketWrapper:
         self.num_sockets = 0
         self.sock_q_pos = 0
         self.unused_sockets = set()
-        self.sockets_read_event = asyncio.Event()
-        self.executor = ThreadPoolExecutor()
+        self.socket_ready_event = asyncio.Event()
+        self.executor = None  # use default executor
         if native or socket_spec.bind or disable_loadbalance:
             self.socket = self.get_socket(socket_spec)
         else:
-            self.init_dns_sockets(socket_spec)
+            self.init_dns_sockets()
 
     def init_socket_q(self):
         self.socket_q = list(self.socket_pool.values())
         self.num_sockets = len(self.socket_q)
         if self.num_sockets > 0:
-            self.sock_q_pos = random.randint(0, self.num_sockets -1)
-            self.sockets_read_event.set()
+            self.sock_q_pos = random.randint(0, self.num_sockets - 1)
+            self.socket_ready_event.set()
         else:
-            self.sockets_read_event.clear()
+            self.socket_ready_event.clear()
 
-    def init_dns_sockets(self, socket_spec: Socket):
-        hosts = get_dns_hosts(socket_spec)
+    def init_dns_sockets(self):
+        hosts = get_dns_hosts(self.socket_spec)
         for ip in hosts:
             self.socket_pool[ip] = self.get_socket(self.socket_for_host(ip))
 
         self.init_socket_q()
 
+    async def check_connection(self) -> bool:
+        if self.native:
+            path = self.socket_spec.get_conn_string(self.native).replace('ipc://', '')
+            return os.path.exists(path)
+        loop = asyncio.get_event_loop()
+        new_dns = await loop.run_in_executor(self.executor, partial(get_dns_hosts, self.socket_spec))
+        return len(new_dns) > 0
+
     def socket_for_host(self, host: str) -> Socket:
         return Socket(self.socket_spec.bind, self.socket_spec.sock_type, self.socket_spec.port, host=host)
 
@@ -128,13 +136,13 @@ class SocketWrapper:
         if self.socket:
             self.socket.send(data)
         else:
-            await self.sockets_read_event.wait()
+            await self.socket_ready_event.wait()
             await self.next_socket.send(data)
 
     async def recv(self) -> bytes:
         if self.socket:
             return await self.socket.recv()
-        await self.sockets_read_event.wait()
+        await self.socket_ready_event.wait()
         recv_socket = self.next_socket
         return await recv_socket.recv()
 
