@@ -9,8 +9,7 @@ from pinecone.protos import core_pb2
 from pinecone.network.zmq.servlet import ZMQServlet
 from pinecone.utils import open_or_create, get_hostname, wal, get_current_namespace, index_state, format_error_msg
 from pinecone.utils.exceptions import RestartException
-from pinecone.utils.constants import MAX_CLIENTS, ZMQ_LOG_PORT, MAX_MSGS_PENDING, KUBE_SYNC_TTL, \
-    ZMQ_PORT_IN, IndexState, COMPACT_WAL_THRESHOLD
+from pinecone.utils.constants import ZMQ_LOG_PORT, KUBE_SYNC_TTL, ZMQ_PORT_IN, IndexState, COMPACT_WAL_THRESHOLD
 from pinecone.network.zmq.spec import ServletSpec, SocketType, Socket
 from pinecone.network.zmq.socket_wrapper import SocketWrapper
 
@@ -22,7 +21,6 @@ from threading import Lock
 import asyncio
 import os
 import functools
-import mmap
 import requests
 
 
@@ -34,10 +32,9 @@ class WALServlet(ZMQServlet):
     - leader-follower replication
     """
 
-    def __init__(self, servlet_spec: ServletSpec, num_replicas: int, path: str, volume_request: int = 15):
+    def __init__(self, servlet_spec: ServletSpec, path: str, volume_request: int = 15):
         super().__init__(servlet_spec)
         self.replica = servlet_spec.replica
-        self.num_replicas = num_replicas
         self.wal_size_limit = volume_request * (10 ** 9) * COMPACT_WAL_THRESHOLD  # translate to GB
 
         self.executor = concurrent.futures.ThreadPoolExecutor()
@@ -45,9 +42,6 @@ class WALServlet(ZMQServlet):
         self.log_in = self.init_socket(self.log_in_socket)
         self.log_out = self.init_socket(self.log_out_socket)
 
-        self.in_buffer = asyncio.Queue()
-        self.in_msgs = asyncio.Queue(maxsize=MAX_MSGS_PENDING)
-
         os.makedirs(path, exist_ok=True)
 
         self.log_path = os.path.join(path, 'log')
@@ -60,16 +54,7 @@ class WALServlet(ZMQServlet):
         self.last = core_pb2.Request()
         self.size = self.load_size()
 
-        offset_path = os.path.join(path, 'offsets')
-        self.offset_file = open_or_create(offset_path, truncate=wal.OFFSET_BYTES*self.num_replicas)
-        self.offsets = mmap.mmap(self.offset_file.fileno(), 0)
-        self.offset_locks = [Lock() for _ in range(0, self.num_replicas)]
-
-        client_offsets_path = os.path.join(path, 'client_offsets')
-        self.client_offsets_file = open_or_create(client_offsets_path, truncate=wal.OFFSET_BYTES*MAX_CLIENTS)
-
-        self.client_offsets = mmap.mmap(self.client_offsets_file.fileno(), 0)
-        self.client_offsets_locks = [Lock() for _ in range(0, MAX_CLIENTS)]
+        self.acked_offset = 0  # wal offset acked
 
         self.hash_ring = ConsistentHashRing.from_file(servlet_spec.function_name, servlet_spec.native)
         self.rebalance_requested = False
@@ -100,7 +85,6 @@ class WALServlet(ZMQServlet):
         return size
 
     def cleanup(self):
-        self.offset_file.close()
         self.log_file.close()
         super().cleanup()
 
@@ -147,7 +131,7 @@ class WALServlet(ZMQServlet):
         while True:
             msg = await self.recv_msg(sock)
             if self.loading_from_wal:
-                await asyncio.sleep(msg.timeout)
+                await asyncio.sleep(msg.timeout * 0.75)
                 if self.loading_from_wal:
                     await self.index_busy(msg)
                     continue
@@ -156,12 +140,15 @@ class WALServlet(ZMQServlet):
                 msg.num_shards = self.hash_ring.size
 
             if self.use_wal(msg):
-                await self.in_msgs.put(msg)
+                await self.handle_msg_wal(msg)
             else:
-                await self.index_ready.wait()
                 loop.create_task(self.handle_msg(msg))
 
-    async def reload_hash_ring(self):
+    async def kube_sync_task(self):
+        """
+        Dynamically relaods number of shards and number of replicas from the control plane
+        :return:
+        """
         loop = asyncio.get_event_loop()
         while True:
             new_hr = await loop.run_in_executor(self.executor, ConsistentHashRing.from_file, self.spec.function_name,
@@ -172,11 +159,12 @@ class WALServlet(ZMQServlet):
 
                 await self.start_compact_log(rebalance=True)
                 self.rebalance_requested = False
+            await self.check_for_new_replicas()
             await asyncio.sleep(KUBE_SYNC_TTL)
 
     def start_reload_task(self) -> List[asyncio.Task]:
         loop = asyncio.get_event_loop()
-        return [*super().start_reload_task(), loop.create_task(self.reload_hash_ring())]
+        return [*super().start_reload_task(), loop.create_task(self.kube_sync_task())]
 
     def start_polling(self) -> List[asyncio.Task]:
         loop = asyncio.get_event_loop()
@@ -186,7 +174,19 @@ class WALServlet(ZMQServlet):
         req_type = msg.WhichOneof('body')
         return req_type in ['index', 'delete']
 
+    async def handle_write(self, msg: core_pb2.Request):
+        self.index_ready.clear()
+        await self.wait_handlers_done()
+        await self.handle_msg(msg)
+        self.index_ready.set()
+
     async def handle_msg_wal(self, msg: core_pb2.Request):
+        """
+        Leader handles write messages. Checks that they won't fill up index, optionally triggers rebalance,
+        and then acknowledges them / sends to replicas
+        :param msg:
+        :return:
+        """
         msg_type = msg.WhichOneof('body')
         if self.hash_ring and msg_type == 'index':
             shard_msg = self.hash_ring.shard_msg(msg)
@@ -195,12 +195,11 @@ class WALServlet(ZMQServlet):
                     await self.forward_shard_msg(msg, shard)
             msg = shard_msg[self.spec.shard]
         elif self.rebalance_requested:
-            format_error_msg(IndexBusyException("Index rebalancing, try again in a minute"), self.spec.function_name)
+            format_error_msg(msg, IndexBusyException("Index rebalancing, try again shortly"), self.spec.function_name)
             await self.send_msg(msg)
 
         new_offset = await self.put(msg)
-        client_id = msg.client_id
-        client_offset = msg.client_offset
+
         idx_state = await self.run_sync(index_state)
         if msg_type == 'index':
             if idx_state == IndexState.FULL:
@@ -214,44 +213,30 @@ class WALServlet(ZMQServlet):
                 await self.run_sync(self.trigger_rebalance_shards)
                 self.rebalance_requested = True
 
-        self.index_ready.clear()
-        await self.wait_handlers_done()
-        await self.handle_msg(msg)
-        self.index_ready.set()
-        await self.set_client_offset(client_id, client_offset + 1)
+        await self.handle_write(msg)
         await self.ack(new_offset)
 
+    async def start_replay_to_replica(self, replica: int, offset: int):
+        event_loop = asyncio.get_event_loop()
+        await self.run_sync(self.handle_replay_request, event_loop, replica, offset, end_offset=self.acked_offset)
+
     async def handle_log_msg(self, msg: 'core_pb2.LogEntry'):
+        """
+        Handles replay messages from replicas to leader, and log entries being sent from leader to replicas
+        :param msg:
+        :return:
+        """
         if not self.leader:
             try:
                 if await self.put(msg.entry, offset=msg.offset):
-                    self.index_ready.clear()
-                    await self.wait_handlers_done()
-                    await self.handle_msg(msg.entry)
-                    self.index_ready.set()
+                    await self.handle_write(msg.entry)
+                    await self.ack(offset=msg.offset)
             except InvalidOffsetException:
                 logger.warning(f"Received invalid offset {msg.offset}")
-                await self.send_log_msg(self.get_replay_request(), 0)
-        else:
-            await self.remote_ack(msg.offset, msg.ack.replica, msg.ack.replay)
-            del msg
-
-    async def consume_msgs(self):
-        while True:
-            msg = await self.in_msgs.get()
-            if self.is_next(msg.client_id, msg.client_offset):
-                await self.handle_msg_wal(msg)
-                while True:
-                    try:
-                        await self.in_msgs.put(self.in_buffer.get_nowait())
-                    except asyncio.QueueEmpty:
-                        break
-            else:
-                # logger.error(f"Invalid client_offset {msg.client_offset}")
-                if msg.client_offset >= self.get_client_offset(msg.client_id):
-                    await self.in_buffer.put(msg)
-                # elif msg.client_offset == self.get_client_offset(msg.client_id) - 1:
-                #     await self.send_msg(msg)
+                await self.send_log_msg(self.get_replay_request(), self.acked_offset)
+        elif msg.ack.replay:
+            await self.start_replay_to_replica(msg.ack.replica, msg.offset)
+        del msg
 
     async def recv_log_msg(self):
         msg = await self.log_in.socket().recv()
@@ -267,51 +252,6 @@ class WALServlet(ZMQServlet):
         loop = asyncio.get_event_loop()
         return await loop.run_in_executor(self.executor, functools.partial(func, *args, **kwargs))
 
-    def get_offset(self, replica: int):
-        start = wal.OFFSET_BYTES * replica
-        return int.from_bytes(self.offsets[start:start + wal.OFFSET_BYTES], wal.BYTEORDER)
-
-    def offsets_list(self):
-        return [self.get_offset(replica) for replica in range(0, self.num_replicas)]
-
-    def write_offset(self, replica: int, offset: int):
-        with self.offset_locks[replica]:
-            self.offsets.seek(replica * wal.OFFSET_BYTES)
-            self.offsets.write(offset.to_bytes(wal.OFFSET_BYTES, wal.BYTEORDER))
-            self.offsets.flush()
-
-    async def set_offset(self, replica: int, offset: int):
-        await self.run_sync(self.write_offset, replica, offset)
-
-    def get_client_offset(self, client_id: int):
-        if client_id >= MAX_CLIENTS:
-            return 0
-        with self.client_offsets_locks[client_id]:
-            start = client_id * wal.OFFSET_BYTES
-            return int.from_bytes(self.client_offsets[start:start + wal.OFFSET_BYTES], wal.BYTEORDER)
-
-    def is_next(self, client_id: int, client_offset: int):
-        return True
-        # if self.leader:
-        #     if client_offset == 0:
-        #         self.sync_set_client_offset(client_id, 0)  # should make this async
-        #     return client_offset == self.get_client_offset(client_id)
-        # else:
-        #     return True
-
-    def sync_set_client_offset(self, client_id: int, client_offset: int):
-        if client_id >= MAX_CLIENTS:
-            return
-        with self.client_offsets_locks[client_id]:
-            start = client_id * wal.OFFSET_BYTES
-            self.client_offsets[start:start + wal.OFFSET_BYTES] = client_offset.to_bytes(wal.OFFSET_BYTES, wal.BYTEORDER)
-            self.client_offsets.flush()
-
-    async def set_client_offset(self, client_id: int, client_offset: int):
-        if self.get_client_offset(client_id) > client_offset:
-            return
-        await self.run_sync(self.sync_set_client_offset, client_id, client_offset)
-
     def sync_put(self, entry: core_pb2.Request, offset: int):
         if (not offset) or (self.size + 1) == offset:
             with self.log_lock:
@@ -322,6 +262,7 @@ class WALServlet(ZMQServlet):
                 logfile_size = self.log_file.tell()
             return logfile_size, self.size
 
+        # ignore because the msg has already been added
         elif offset <= self.size:
             return self.log_file.tell(), None
         else:
@@ -337,6 +278,17 @@ class WALServlet(ZMQServlet):
             loop.create_task(self.start_compact_log())
         return offset
 
+    async def check_for_new_replicas(self):
+        old_max_replica = self.log_out.max_replica
+        max_replica = await self.log_out.get_max_replica(self.spec.shard)
+        logger.success(f"Max replica: {max_replica} old: {old_max_replica}")
+
+        msg = core_pb2.LogEntry(entry=self.last, offset=self.acked_offset)
+        serialized_msg = msg.SerializeToString()
+        for replica in range(old_max_replica + 1, max_replica + 1):
+            logger.info(f"New replica {replica} detected, replaying WAL")
+            await self.send_log_msg(serialized_msg, replica)
+
     def sync_replay_log(self, offset: int, end_offset: int = None) -> Iterable[Tuple[int, core_pb2.Request]]:
         """
         Replay log from offset 0 (inclusive) to end_offset (exclusive)
@@ -344,11 +296,11 @@ class WALServlet(ZMQServlet):
         :param end_offset:
         :return:
         """
-        with self.log_lock:
-            self.log_file.seek(0)
+        with open_or_create(self.log_path) as log_file:
+            log_file.seek(0)
             pos = 0
             while True:
-                entry, msg_size = wal.read_entry(self.log_file)
+                entry, msg_size = wal.read_entry(log_file)
 
                 if msg_size == 0 or (end_offset and pos >= end_offset):
                     break
@@ -357,7 +309,6 @@ class WALServlet(ZMQServlet):
                 pos += 1
 
     async def forward_shard_msg(self, msg: core_pb2.Request, shard: int):
-        msg.client_offset = 0
         await self.rebalance_sock.send_msg(self.rebalance_sock.socket(shard, replica=0), msg)
 
     def handle_replay_request(self, loop: asyncio.AbstractEventLoop, replica: int, offset: int, end_offset: int = None):
@@ -366,69 +317,52 @@ class WALServlet(ZMQServlet):
             asyncio.run_coroutine_threadsafe(self.send_log_msg(core_pb2.LogEntry(entry=entry,
                                                                                  offset=pos), replica), loop)
 
-    def get_leader_log_backlog(self, replica: int) -> Iterable[core_pb2.LogEntry]:
-        return [core_pb2.LogEntry(entry=entry, offset=offset)
-                for offset, entry in self.sync_replay_log(self.get_offset(replica))]
-
-    def sync_ack(self, offset: int):
-        self.write_offset(self.replica, offset)
-
     async def ack(self, offset: int):
-        prev_offset = self.get_offset(self.replica)
-        if prev_offset > offset:
+        if self.acked_offset > offset:
             logger.warning('Acked offset less than current offset')
-        # elif prev_offset + 1 > offset:
-        #     logger.warning('Acked by more than one offset')
+        elif self.acked_offset + 1 > offset:
+            logger.warning('Acked by more than one offset')
 
-        await self.set_offset(self.replica, offset)
+        self.acked_offset = offset
         if self.leader:
             msg = core_pb2.LogEntry(entry=self.last, offset=offset)
             serialized_msg = msg.SerializeToString()
-            for r in range(1, self.num_replicas):
+
+            for r in range(1, self.log_out.max_replica + 1):
                 await self.send_log_msg(serialized_msg, r)
 
-        else:
-            ack = core_pb2.Ack(replica=self.replica)
-            await self.send_log_msg(core_pb2.LogEntry(offset=offset, ack=ack), 0)
-
-    async def remote_ack(self, start_offset: int, replica: int, replay=False):
-        if replay:
-            leader_offset = self.get_offset(self.replica)
-            event_loop = asyncio.get_event_loop()
-            await self.run_sync(self.handle_replay_request, event_loop, replica, start_offset, end_offset=leader_offset)
-
-    def get_replay_request(self, override_offset=None):
+    def get_replay_request(self,):
         ack = core_pb2.Ack(replica=self.replica, replay=True)
-        offset = override_offset or self.get_offset(self.replica)
-        logger.info(f"Sending replay request for offset {offset}")
-        return core_pb2.LogEntry(offset=offset, ack=ack)
+        logger.info(f"Sending replay request for offset {self.acked_offset}")
+        return core_pb2.LogEntry(offset=self.acked_offset, ack=ack)
 
     def load_sync(self, handler):
-        logger.info(f'Recovering from local WAL: offset: {self.get_offset(self.replica)}')
+        logger.info(f'Recovering from local WAL: offset: {self.acked_offset}')
         self.loading_from_wal = True
-        for offset, entry in self.sync_replay_log(self.get_offset(self.replica)):
+        for offset, entry in self.sync_replay_log(self.acked_offset):
             try:
                 handler(entry)
             except Exception as e:  # ignore messages that fail
                 logger.info(f'Error handling msg while recovering from WAL. offset: {offset} error: {e}')
-            self.sync_ack(offset)
+            self.acked_offset = offset
         self.loading_from_wal = False
 
     def swap_log_files(self, end_file_offset: int):
         tmp_log_file = open(self.tmp_log_path, "r+b")
 
-        with self.log_lock:
-            self.log_file.seek(end_file_offset)
+        with open_or_create(self.log_path) as log_file:
+            log_file.seek(end_file_offset)
             while True:
-                entry, msg_size = wal.read_entry(self.log_file)
+                entry, msg_size = wal.read_entry(log_file)
                 if msg_size == 0:
                     break
                 wal.write_log_entry(tmp_log_file, entry)
-            tmp_log_file.close()
-            self.log_file.close()
-            os.remove(self.log_path)
-            os.rename(self.tmp_log_path, self.log_path)
-            self.log_file = open(self.log_path, "r+b")
+
+        tmp_log_file.close()
+        self.log_file.close()
+        os.remove(self.log_path)
+        os.rename(self.tmp_log_path, self.log_path)
+        self.log_file = open(self.log_path, "r+b")
 
     async def start_compact_log(self, rebalance=False):
         while self.log_compaction_started:
@@ -443,7 +377,7 @@ class WALServlet(ZMQServlet):
         loop = asyncio.get_event_loop()
         await loop.run_in_executor(concurrent.futures.ProcessPoolExecutor(), wal.compact_log,
                                    self.log_path, self.tmp_log_path, end_file_offset, self.hash_ring, self.spec.shard,
-                                   self._rebalance_sock_spec, self.spec.native)
+                                   self._rebalance_sock_spec if self.leader else None, self.spec.native)
 
         logger.info("Starting swap log files...")
         await self.run_sync(self.swap_log_files, end_file_offset)
